//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by assimp_view.rc
//
#define IDC_MYICON                      2
#define IDD_ASSIMP_VIEW_DIALOG          102
#define IDD_ABOUTBOX                    103
#define IDI_ASSIMP_VIEW                 107
#define IDI_SMALL                       108
#define IDR_MAINFRAME                   128
#define IDD_DIALOGMAIN                  129
#define IDB_BITMAP1                     130
#define IDR_MENU1                       131
#define IDD_LOADDIALOG                  132
#define IDB_BITMAP2                     134
#define IDD_AVHELP                      135
#define IDR_TEXT1                       138
#define IDR_MAX                         140
#define IDB_MAXCIRCLEMASK               141
#define IDB_MAXCIRCLE                   142
#define IDR_HUD                         143
#define IDR_HUDMASK                     144
#define IDB_BANIM                       145
#define IDB_BITMAP4                     146
#define IDB_BDISPLAY                    146
#define IDB_BITMAP5                     147
#define IDB_BINTER                      147
#define IDB_BITMAP6                     148
#define IDB_BRENDERING                  148
#define IDB_BITMAP7                     149
#define IDB_BSTATS                      149
#define IDB_BTX                         150
#define IDB_BITMAP8                     151
#define IDB_BFX                         151
#define IDB_BITMAP9                     152
#define IDB_BNODE                       152
#define IDB_BITMAP10                    153
#define IDB_BROOT                       153
#define IDD_LOGVIEW                     154
#define IDB_BTXI                        155
#define IDR_TXPOPUP                     156
#define IDR_MATPOPUP                    157
#define IDD_DIALOGSMOOTH                159
#define SVNRevision                     700
#define IDC_CHECK1                      1000
#define IDC_TOGGLEMS                    1000
#define IDC_CHECK2                      1001
#define IDC_TOGGLEWIRE                  1001
#define IDC_CHECK3                      1002
#define IDC_TOGGLEMAT                   1002
#define IDC_CHECK4                      1003
#define IDC_TOGGLENORMALS               1003
#define IDC_CHECK5                      1004
#define IDC_AUTOROTATE                  1004
#define IDC_RT                          1006
#define IDC_NUMVERTS                    1007
#define IDC_NUMFACES                    1008
#define IDC_NUMMATS                     1009
#define IDC_FPS                         1010
#define IDC_EFPS                        1011
#define IDC_EMAT                        1012
#define IDC_EFACE                       1013
#define IDC_EVERT                       1014
#define IDC_CHECK6                      1015
#define IDC_LIGHTROTATE                 1015
#define IDC_3LIGHTS                     1016
#define IDC_PROGRESS                    1016
#define IDC_LOADTIME                    1017
#define IDC_ELOAD                       1018
#define IDC_CHECK7                      1019
#define IDC_ZOOM                        1019
#define IDC_CHECK8                      1020
#define IDC_LOWQUALITY                  1020
#define IDC_NUMMATS2                    1021
#define IDC_NUMSHADERS                  1021
#define IDC_ESHADER                     1022
#define IDC_RICHEDIT21                  1023
#define IDC_EMESH                       1023
#define IDC_CHECK9                      1024
#define IDC_NOSPECULAR                  1024
#define IDC_PLAYANIM                    1025
#define IDC_3LIGHTS2                    1025
#define IDC_NOAB                        1025
#define IDC_SPEED                       1026
#define IDC_COMBO1                      1027
#define IDC_PINORDER                    1028
#define IDC_NOSPECULAR2                 1028
#define IDC_SSPEED                      1029
#define IDC_SANIM                       1030
#define IDC_SANIMGB                     1031
#define IDC_ENODE                       1031
#define IDC_ESHADER2                    1032
#define IDC_ETEX                        1032
#define IDC_TREE1                       1033
#define IDC_EDIT1                       1034
#define IDC_BLUBB                       1037
#define IDC_BLABLA                      1038
#define IDC_NUMNODES                    1038
#define IDC_LCOLOR1                     1041
#define IDC_LCOLOR2                     1042
#define IDC_ENODEWND                    1043
#define IDC_LCOLOR3                     1044
#define IDC_LRESET                      1046
#define IDC_NUMMESHES                   1047
#define IDC_VIEWMAT                     1048
#define IDC_VIEWMATRIX                  1048
#define IDC_SLIDERANIM                  1052
#define IDC_PLAY                        1053
#define IDC_SHOWSKELETON                1054
#define IDC_BFCULL                      1055
#define IDC_EDITSM                      1056
#define ID_VIEWER_OPEN                  32771
#define ID_VIEWER_CLOSETHIS             32772
#define ID_VIEWER_CLOSEASSET            32773
#define ID_VIEWER_QUIT                  32774
#define ID__ABOUT                       32775
#define ID__HELP                        32776
#define ID_VIEWER_SAVESCREENSHOTTOFILE  32777
#define ID_VIEWER_RESETVIEW             32778
#define ID_BACKGROUND_LOADTEXTURE       32779
#define ID_BACKGROUND_CLEAR             32780
#define ID_BACKGROUND_SETCOLOR          32781
#define ID_Menu                         32782
#define ID_BACKGROUND_LOADSKYBOX        32783
#define ID_VIEWER_H                     32784
#define ID_TOOLS_LOGWINDOW              32785
#define ID_TOOLS_SAVELOGTOFILE          32786
#define ID_TOOLS_CLEARLOG               32787
#define ID_VIEWER_RECENTFILES           32788
#define ID_VIEWER_MEMORYCONSUMATION     32789
#define ID_VIEWER_CLEARHISTORY          32790
#define ID_TOOLS_ORIGINALNORMALS        32791
#define ID_TOOLS_SMOOTHNORMALS          32792
#define ID_TOOLS_HARDNORMALS            32793
#define ID_TOOLS_FLIPNORMALS            32794
#define ID__S                           32795
#define ID__FEEDBACK                    32796
#define ID_FEEDBACK_GH                  32797
#define ID_FEEDBACK_FEATUREREQUEST      32798
#define ID_FEEDBACK_DONATE              32799
#define ID_ANIMATION_PLAYALLINORDER     32800
#define ID_TOOLS_STEREOVIEW             32801
#define ID_EGNEKLGEG_EGEG               32802
#define ID_HEY_REPLACE                  32803
#define ID_HEY_EXPORT                   32804
#define ID_HEY_REMOVE                   32805
#define ID_SOLONG_ADDDIFFUSETEXTURE     32806
#define ID_SOLONG_ADDSPECULARTEXTURE    32807
#define ID_SOLONG_ADDAMBIENTTEXTURE     32808
#define ID_SOLONG_ADDEMISSIVETEXTURE    32809
#define ID_SOLONG_ADDOPACITYTEXTURE     32810
#define ID_SOLONG_ADDNORMAL             32811
#define ID_SOLONG_ADDSHININESSTEXTURE   32812
#define ID_SOLONG_CLEARDIFFUSECOLOR     32813
#define ID_SOLONG_CLEARSPECULARCOLOR    32814
#define ID_SOLONG_CLEARAMBIENTCOLOR     32815
#define ID_SOLONG_CLEAREMISSIVECOLOR    32816
#define ID_SOLONG_CLEARTRANSPARENCY     32817
#define ID_SOLONG_MAKEDEFAULTMATERIAL   32818
#define ID_HEY_RESETTEXTURE             32819
#define ID_SOLONG_SETSHADINGMODE        32820
#define ID_SETSHADINGMODE_GOURAUD       32821
#define ID_SETSHADINGMODE_PHONG         32822
#define ID_OPTIMIZE_OPTIMIZEACMR        32823
#define ID_OPTIMIZE_OPTIMIZEOVERDRAW    32824
#define ID_OPTIMIZE_OPTIMIZEBOTH        32825
#define ID_VERTEXCACHELOCALITY_FINDCURRENT 32826
#define ID_VERTEXCACHELOCALITY_OPTIMIZE 32827
#define ID_VERTEXCACHELOCALITY_FINDBEST 32828
#define ID_OPTIMIZE_SCENEGRAPH          32829
#define ID_SCENEGRAPH_SMALLESTPOSSIBLEGRAPH 32830
#define ID_SMOOTHNORMALS_5              32831
#define ID_SMOOTHNORMALS_6              32832
#define ID_SMOOTHNORMALS_MAXANGLE60     32833
#define ID_SMOOTHNORMALS_MAXANGLE90     32834
#define ID_SMOOTHNORMALS_MAXANGLE120    32835
#define ID_SMOOTHNORMALS_NOLIMIT        32836
#define ID_SMOOTHNORMALS_30             32837
#define ID_SMOOTHNORMALS_40             32838
#define ID_SMOOTHNORMALS_60             32839
#define ID_SMOOTHNORMALS_90             32840
#define ID_SMOOTHNORMALS_120            32841
#define ID_Menu32842                    32842
#define ID_SMOOTHANGLE_30               32843
#define ID_SMOOTHANGLE_40               32844
#define ID_SMOOTHANGLE_50               32845
#define ID_SMOOTHANGLE_60               32846
#define ID_SMOOTHANGLE_70               32847
#define ID_SMOOTHANGLE_90               32848
#define ID_SMOOTHANGLE_120              32849
#define ID_SMOOTHANGLE_NONE             32850
#define ID_TOOLS_SETANGLELIMIT          32851
#define ID_VIEWER_PP_JIV                32852
#define ID_VIEWER_PP_RRM                32852
#define ID_VIEWER_PP_OM                 32853
#define ID_VIEWER_PP_OG                 32854
#define ID_VIEWER_PP_FIM                32855
#define ID_VIEWER_PP_VDS                32856
#define ID_VIEWER_PP_PTV                32857
#define ID_VIEWER_PP_ICL                32858
#define ID_VIEWER_PP_FIN                32859
#define ID_VIEWER_PP_FD                 32860
#define ID_VIEWER_PP_FID                32861
#define ID_VIEWER_PP_GUV                32862
#define ID_VIEWER_PP_TUV                32863
#define ID_VIEWER_PP_RLINE_PNT          32864
#define ID_REPORTBUG                    32865
#define ID_FR                           32866
#define ID__WEBSITE                     32867
#define ID__SF                          32868
#define ID__                            32869
#define ID__WEBSITESF                   32870
#define ID_IMPORTSETTINGS_CALCULATETANGENTSPACE 32871
#define ID_VIEWER_CTS                   32872
#define ID_VIEWER_PP_CTS                32873
#define ID_VIEWER_RELOAD                32874
#define ID_VIEWER_PP_RRM2               32875
#define ID_IMPORTSETTINGS_RESETTODEFAULT 32876
#define ID_IMPORTSETTINGS_OPENPOST      32877
#define ID_EXPORT                       32878
#define ID_IMPORTSETTINGS_REMOVEDUMMYBONES 32879
#define ID_VIEWER_PP_DB                 32880
#define IDC_STATIC                      -1

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC                     1
#define _APS_NEXT_RESOURCE_VALUE        160
#define _APS_NEXT_COMMAND_VALUE         32881
#define _APS_NEXT_CONTROL_VALUE         1059
#define _APS_NEXT_SYMED_VALUE           110
#endif
#endif
