unit aiMatrix4x4;

interface

type TaiMatrix4x4 = packed record
   a1, a2, a3, a4: single;
   b1, b2, b3, b4: single;
   c1, c2, c3, c4: single;
   d1, d2, d3, d4: single;
end;
PaiMatrix4x4 = ^TaiMatrix4x4;


implementation

end.
