client_actions = [
    'w',     # Move forward    
    'a',     # Move left
    's',     # Move backward
    'd',     # Move right
    'z',     # Move downward
    'x',    # Move upward
    'j',    # Turn left
    'l'     # Turn right
]



client_newloc  = [
    'n'        # Bring the view to a new starting location
]